import getpass
import os
import sys
import errno
import signal
import select
import termios
import tty

from typing import Dict
from typing import List
from typing import Tuple
from typing import Optional
from typing import Iterable
from . import px_process
from . import px_sort_order


# NOTE: To work with this list it can be useful to find the text "Uncomment to
# debug input characters" in handle_search_keypresses() in px_top.py.
KEY_ESC = "\x1b"
KEY_DELETE = "\x1b[3~"
KEY_BACKSPACE = "\x7f"
KEY_UPARROW = "\x1b[A"
KEY_DOWNARROW = "\x1b[B"
KEY_ENTER = "\x0d"


initial_termios_attr = None


CSI = "\x1b["

CURSOR_TO_TOP_LEFT = CSI + "H"
CURSOR_TO_RIGHT_EDGE = CSI + "999C"  # Actually "999 steps to the right"

"""
Clear the screen and move cursor to top left corner:
https://en.wikipedia.org/wiki/ANSI_escape_code

Note that some experimentation was involved in coming up with this
exact sequence; if you do first "clear the full screen" then "home"
for example, the contents of the previous screen gets added to the
scrollback buffer, which isn't what we want. Tread carefully if you
intend to change these.
"""
CLEAR_SCREEN = CSI + "1J" + CURSOR_TO_TOP_LEFT

HIDE_CURSOR = CSI + "?25l"
SHOW_CURSOR = CSI + "?25h"

CLEAR_TO_EOL = CSI + "0K"
CLEAR_TO_END_OF_SCREEN = CSI + "J"  # Clear from cursor to end of screen

# Used for informing our getch() function that a window resize has occurred
SIGWINCH_PIPE = os.pipe()

# We'll report window resize as this key having been pressed.
#
# NOTE: This must be detected as non-printable by handle_search_keypress().
SIGWINCH_KEY = "\x00"

_enable_color = True

previous_screen_lines: List[str] = []
previous_screen_columns: int = 0


def disable_color():
    global _enable_color
    _enable_color = False


def sigwinch_handler(_signum, _frame):
    """Handle window resize signals by telling our getch() function to return"""
    # "r" for "refresh" perhaps? The actual letter doesn't matter, as long as it
    # doesn't collide with anything with some meaning other than "please
    # redraw".
    os.write(SIGWINCH_PIPE[1], SIGWINCH_KEY.encode("utf-8"))


# Subscribe to window size change signals
signal.signal(signal.SIGWINCH, sigwinch_handler)


class ConsumableString:
    def __init__(self, string: str) -> None:
        self._string = string

    def __len__(self):
        return len(self._string)

    def consume(self, to_consume: str) -> bool:
        if not self._string.startswith(to_consume):
            return False

        self._string = self._string[len(to_consume) :]
        return True


def read_select(
    fds: List[int],
    timeout_seconds: Optional[int] = None,
) -> List[int]:
    """Select on any of the fds becoming ready for read, retry on EINTR"""

    # NOTE: If you change this method, you must test running "px --top" and
    # resize the window
    while True:
        try:
            return select.select(fds, [], [], timeout_seconds)[0]
        except OSError as ose:
            if ose.errno == errno.EINTR:
                # EINTR happens when the terminal window is resized by the user,
                # just try again.
                continue

            # Non-EINTR exceptions are unexpected, help!
            raise


def getch(
    timeout_seconds: Optional[int] = None, fd: Optional[int] = None
) -> Optional[ConsumableString]:
    """
    Wait at most timeout_seconds for a character to become available on stdin.

    Returns the character, or None on timeout.
    """
    if fd is None:
        fd = sys.stdin.fileno()

    can_read_from = read_select([fd, SIGWINCH_PIPE[0]], timeout_seconds)

    # Read all(ish) bytes from the first ready-for-read stream. If more than one
    # stream is ready, we'll catch the second one on the next call to this
    # function, so just doing the first is fine.
    for stream in can_read_from:
        string = os.read(stream, 1234).decode("UTF-8")
        if len(string) > 0:
            return ConsumableString(string)

        # A zero length response means we got EOF from one of the streams. This
        # happens (at least) during testing.
        continue

    return None


class TerminalError(Exception):
    pass


def get_window_size() -> Tuple[int, int]:
    """
    Return the terminal window size as tuple (rows, columns).

    Raises a TerminalError if the window size is unavailable.
    """
    if not sys.stdout.isatty():
        # We shouldn't truncate lines when piping
        raise TerminalError("Not a TTY, window size not available")

    result = None
    with os.popen("stty size", "r") as stty_size:
        result = stty_size.read().split()
    if len(result) < 2:
        # Getting the terminal window width failed, don't truncate
        raise TerminalError("Incomplete response to window size query")

    columns = int(result[1])
    if columns < 1:
        # This seems to happen during OS X CI runs:
        # https://travis-ci.org/walles/px/jobs/113134994
        raise TerminalError("Window width unreasonably small: " + str(columns))

    rows = int(result[0])
    if rows < 1:
        # Don't know if this actually happens, we just do it for symmetry with
        # the columns check above
        raise TerminalError("Window height unreasonably small: " + str(rows))

    return (rows, columns)


def raw_lines_to_screen_lines(raw_lines: List[str], columns: int) -> List[str]:
    """Crop the lines to the right length and add clear-to-EOL at the end"""
    screen_lines: List[str] = []

    for raw_line in raw_lines:
        cooked = crop_ansi_string_at_length(raw_line, columns)
        cooked += CLEAR_TO_EOL
        screen_lines.append(cooked)

    return screen_lines


def filter_out_unchanged_screen_lines(
    all_screen_lines: List[str], columns: int
) -> List[Optional[str]]:
    """Compare lines to the cache and None out the ones that match the cache"""

    if len(previous_screen_lines) != len(all_screen_lines):
        # Screen resized, never mind the cache
        return all_screen_lines  # type: ignore
    if previous_screen_columns != columns:
        # Screen resized, never mind the cache
        return all_screen_lines  # type: ignore

    filtered: List[Optional[str]] = []
    for line_index, line in enumerate(all_screen_lines):
        line_from_previous_screen = previous_screen_lines[line_index]
        if line == line_from_previous_screen:
            # Cache hit, don't draw this line
            filtered.append(None)
        else:
            filtered.append(line)

    return filtered


def draw_screen_lines(lines: List[str], columns: int) -> None:
    unfiltered_screen_lines = raw_lines_to_screen_lines(lines, columns)
    screen_lines = filter_out_unchanged_screen_lines(unfiltered_screen_lines, columns)

    screenstring = CURSOR_TO_TOP_LEFT
    skip_lines = 0
    for screen_line in screen_lines:
        if screen_line is None:
            skip_lines += 1
        else:
            if skip_lines > 0:
                # Move down the required number of lines
                screenstring += CSI + str(skip_lines) + "E"
                skip_lines = 0

            # The line changed, update it!
            screenstring += screen_line

            # Move to next line
            skip_lines = 1

    if skip_lines > 1:
        # Move down the required number of lines. "- 1" because if we skip after
        # the last line, there'll be a leftover line when we enter the
        # per-process menu.
        screenstring += CSI + str(skip_lines - 1) + "E"
        skip_lines = 0

    # Keep the cache up to date
    global previous_screen_lines
    global previous_screen_columns
    previous_screen_lines = unfiltered_screen_lines
    previous_screen_columns = columns

    # In case we got fewer lines than what fits on screen, clear the rest of it.
    # We must start at the right edge of screen doing this in case differential
    # rendering made us not do anything for the last screen line.
    screenstring += CURSOR_TO_RIGHT_EDGE + CLEAR_TO_END_OF_SCREEN

    os.write(sys.stdout.fileno(), screenstring.encode("utf-8"))


def width_specifier(width: int, right_align: bool = False) -> str:
    return_me = "{:"

    if right_align:
        return_me += ">"

    return_me += str(width)
    return_me += "}"

    return return_me


def format_with_widths(widths: List[int], strings: List[str]) -> str:
    """
    Put strings next to each other with a space between each.

    Negative column widths means right aligned text.

    Zero width columns are unlimited.
    """

    assert len(widths) == len(strings)

    result = ""
    for i, width in enumerate(widths):
        if i > 0:
            result += " "

        string = strings[i]

        if width == 0:
            result += string
            continue

        padding_amount = abs(width) - visual_length(string)
        if padding_amount <= 0:
            # NOTE: Future improvement: Crop here rather than just pretending
            # everything is fine
            result += string
            continue

        padding = padding_amount * " "
        if width > 0:
            result += string + padding
        else:
            # Right align
            result += padding + string

    return result


def to_screen_lines(
    procs: List[px_process.PxProcess],
    row_to_highlight: Optional[int],
    sort_order: Optional[px_sort_order.SortOrder],
    with_username: bool = True,
) -> List[str]:
    """
    Returns an array of lines that can be printed to screen. Lines are not
    cropped, so they can be longer than the screen width.

    If sort_order is set, the sort order column will be highlighted.
    """

    cputime_name = "CPUTIME"
    if sort_order == px_sort_order.SortOrder.AGGREGATED_CPU:
        cputime_name = "AGGRCPU"
    headings = [
        "PID",
        "COMMAND",
        "USERNAME",
        "CPU",
        cputime_name,
        "RAM",
        "COMMANDLINE",
    ]
    highlight_column = None
    if sort_order == px_sort_order.SortOrder.MEMORY:
        highlight_column = 5  # "RAM"
    elif sort_order in [
        px_sort_order.SortOrder.CPU,
        px_sort_order.SortOrder.AGGREGATED_CPU,
    ]:
        highlight_column = 4  # "CPUTIME" or "AGGRCPU"

    # Compute widest width for pid, command, user, cpu and memory usage columns
    pid_width = len(headings[0])
    command_width = len(headings[1])
    username_width = len(headings[2])
    cpu_width = len(headings[3])
    cputime_width = len(headings[4])
    mem_width = len(headings[5])
    for proc in procs:
        pid_width = max(pid_width, len(str(proc.pid)))

        # Unindent PIDs 0 and 1. Not fully correct, but since PID 1 has only one
        # parent indenting it doesn't add much information, but does use up
        # space, so let's just not.
        level = max(proc.level - 1, 0)

        command_width = max(command_width, len(proc.command) + level * 2)
        username_width = max(username_width, len(proc.username))
        cpu_width = max(cpu_width, len(proc.cpu_percent_s))

        cputime_s = proc.cpu_time_s
        if sort_order == px_sort_order.SortOrder.AGGREGATED_CPU:
            cputime_s = proc.aggregated_cpu_time_s
        cputime_width = max(cputime_width, len(cputime_s))

        mem_width = max(mem_width, len(proc.memory_percent_s))

    column_widths = [
        -pid_width,
        command_width,
        username_width,
        -cpu_width,
        -cputime_width,
        -mem_width,
        0,  # The command line can have any length
    ]

    username_index = headings.index("USERNAME")
    if not with_username:
        del headings[username_index]
        del column_widths[username_index]

    # Print process list using the computed column widths
    lines = []

    # Highlight the highlight_column
    if highlight_column is not None:
        headings[highlight_column] = underline(headings[highlight_column])

    heading_line = format_with_widths(column_widths, headings)
    heading_line = bold(heading_line)
    lines.append(heading_line)

    max_cpu_percent_s: Optional[str] = None
    max_cpu_percent = 0.0
    max_memory_percent_s: Optional[str] = None
    max_memory_percent = 0.0
    max_cpu_time_seconds = 0.0
    for proc in procs:
        if proc.cpu_percent is not None and proc.cpu_percent > max_cpu_percent:
            max_cpu_percent = proc.cpu_percent
            max_cpu_percent_s = proc.cpu_percent_s
        if proc.memory_percent is not None and proc.memory_percent > max_memory_percent:
            max_memory_percent = proc.memory_percent
            max_memory_percent_s = proc.memory_percent_s

        cpu_time_seconds = proc.cpu_time_seconds
        if sort_order == px_sort_order.SortOrder.AGGREGATED_CPU:
            if proc.pid <= 1:
                # Both the kernel (PID 0) and the init process (PID 1) will just
                # contain the total time of all other processes. Since we only
                # use this max value for highlighting (see below), if we include
                # these only they will be highlighted. So we skip them.
                continue
            cpu_time_seconds = proc.aggregated_cpu_time_seconds
        if cpu_time_seconds is not None and cpu_time_seconds > max_cpu_time_seconds:
            max_cpu_time_seconds = cpu_time_seconds

    current_user = os.environ.get("SUDO_USER") or getpass.getuser()
    for line_number, proc in enumerate(procs):
        cpu_percent_s = proc.cpu_percent_s
        if proc.cpu_percent_s == "0%":
            cpu_percent_s = faint(cpu_percent_s.rjust(cpu_width))
        elif proc.cpu_percent_s == max_cpu_percent_s:
            cpu_percent_s = bold(cpu_percent_s.rjust(cpu_width))

        memory_percent_s = proc.memory_percent_s
        if proc.memory_percent_s == "0%":
            memory_percent_s = faint(memory_percent_s.rjust(mem_width))
        elif proc.memory_percent_s == max_memory_percent_s:
            memory_percent_s = bold(memory_percent_s.rjust(mem_width))

        cpu_time_s = proc.cpu_time_s
        cpu_time_seconds = proc.cpu_time_seconds
        if sort_order == px_sort_order.SortOrder.AGGREGATED_CPU:
            cpu_time_s = proc.aggregated_cpu_time_s
            cpu_time_seconds = proc.aggregated_cpu_time_seconds

        if not cpu_time_seconds:
            # Zero or undefined
            cpu_time_s = faint(cpu_time_s.rjust(cputime_width))
        elif cpu_time_seconds > 0.75 * max_cpu_time_seconds:
            cpu_time_s = bold(cpu_time_s.rjust(cputime_width))
        elif cpu_time_seconds < 0.1 * max_cpu_time_seconds:
            cpu_time_s = faint(cpu_time_s.rjust(cputime_width))

        # NOTE: This logic should match its friend in
        # px_processinfo.py/print_process_tree()
        owner = proc.username
        if owner == "root":
            owner = faint(owner)
        elif owner != current_user:
            # Neither root nor ourselves, highlight!
            owner = bold(owner)

        indent = ""
        if sort_order == px_sort_order.SortOrder.AGGREGATED_CPU:
            # Unindent PIDs 0 and 1. Not fully correct, but since PID 1 has only
            # one parent indenting it doesn't add much information, but does use
            # up space, so let's not.
            level = max(proc.level - 1, 0)
            indent = " " * level * 2

        columns = [
            str(proc.pid),
            indent + proc.command,
            owner,
            cpu_percent_s,
            cpu_time_s,
            memory_percent_s,
            proc.cmdline,
        ]
        if not with_username:
            del columns[username_index]
        line = format_with_widths(column_widths, columns)

        if row_to_highlight == line_number:
            # Highlight the whole screen line
            line = inverse_video(line + " " * 999)
        lines.append(line)

    lines[0] = heading_line

    return lines


def inverse_video(string: str) -> str:
    if not _enable_color:
        return string
    return CSI + "7m" + string + CSI + "27m"


def bold(string: str) -> str:
    if not _enable_color:
        return string
    return CSI + "1m" + string + CSI + "22m"


def faint(string: str) -> str:
    if not _enable_color:
        return string
    return CSI + "2m" + string + CSI + "22m"


def underline(string: str) -> str:
    if not _enable_color:
        return string
    return CSI + "4m" + string + CSI + "24m"


def red(string: str) -> str:
    if not _enable_color:
        return string
    return CSI + "1;97;41m" + string + CSI + "49;39;22m"


def yellow(string: str) -> str:
    if not _enable_color:
        return string
    return CSI + "30;103m" + string + CSI + "49;39m"


def green(string: str) -> str:
    if not _enable_color:
        return string
    return CSI + "1;32m" + string + CSI + "39;22m"


def blue(string: str) -> str:
    if not _enable_color:
        return string
    return CSI + "1;97;44m" + string + CSI + "49;39;22m"


def get_string_of_length(string: str, length: Optional[int]) -> str:
    if length is None:
        return string

    initial_length = visual_length(string)
    if initial_length == length:
        return string

    if initial_length < length:
        return string + " " * (length - initial_length)

    if initial_length > length:
        return crop_ansi_string_at_length(string, length)

    assert False  # How did we end up here?


def _tokenize(string: str) -> Iterable[str]:
    """
    Tokenizes string into character sequences and ANSI sequences.
    """
    i = 0
    while i < len(string):
        try:
            next_csi_index = string.index(CSI, i)
        except ValueError:
            # No more CSIs
            break
        if next_csi_index > i:
            # Yield char sequence until next CSI marker
            yield string[i:next_csi_index]
            i = next_csi_index
            continue

        # We are at a CSI marker
        try:
            next_m_index = string.index("m", i)
        except ValueError:
            # No end-of-escape sequence found
            break
        next_i = next_m_index + 1

        # Yield full escape sequence
        yield string[i:next_i]
        i = next_i
        continue

    # Yield the remaining part of the string
    yield string[i:]


crop_cache: Dict[Tuple[str, int], str] = {}


def crop_ansi_string_at_length(string: str, length: int) -> str:
    assert length >= 0

    cache_key = (string, length)
    cache_hit = crop_cache.get(cache_key, None)
    if cache_hit is not None:
        return cache_hit
    if len(crop_cache) > 400:
        # LRU would be better but this is easier to implement
        crop_cache.clear()

    result = ""
    char_count = 0

    reset_sequence = ""

    for token in _tokenize(string):
        if token.startswith(CSI):
            reset_sequence = CSI + "0m"
            if token == reset_sequence:
                # Already reset
                reset_sequence = ""
            result += token
            continue

        # Not a CSI token
        missing_count = length - char_count
        if len(token) >= missing_count:
            result += token[:missing_count]
            break

        result += token
        char_count += len(token)

    cache_hit = result + reset_sequence
    crop_cache[cache_key] = cache_hit
    return cache_hit


def visual_length(string: str) -> int:
    """
    If we print this string, possibly containing ANSI characters, to
    screen, how many characters wide will it be?
    """
    count = 0
    for token in _tokenize(string):
        if not token.startswith(CSI):
            # These are characters
            count += len(token)

    return count


def _enter_fullscreen():
    global initial_termios_attr
    assert initial_termios_attr is None

    fd = sys.stdin.fileno()

    initial_termios_attr = termios.tcgetattr(sys.stdin.fileno())

    tty.setraw(fd)

    os.write(fd, HIDE_CURSOR.encode("utf-8"))


def _exit_fullscreen():
    global initial_termios_attr
    assert initial_termios_attr is not None

    fd = sys.stdin.fileno()
    os.write(fd, SHOW_CURSOR.encode("utf-8"))

    tty.setcbreak(fd)

    # tty.setraw() disables local echo, so we have to re-enable it here.
    # See the "source code" comment to this answer:
    # http://stackoverflow.com/a/8758047/473672
    if initial_termios_attr:
        termios.tcsetattr(fd, termios.TCSADRAIN, initial_termios_attr)
    initial_termios_attr = None


class fullscreen_display:
    def __enter__(self):
        _enter_fullscreen()

    def __exit__(self, exception_type, exception_value, exception_traceback):
        _exit_fullscreen()

        # Re-raise any exception:
        # https://docs.python.org/2.5/whatsnew/pep-343.html#context-managers
        return False


class normal_display:
    """
    Pause fullscreen mode
    """

    def __enter__(self):
        _exit_fullscreen()

    def __exit__(self, exception_type, exception_value, exception_traceback):
        _enter_fullscreen()

        # Re-raise any exception:
        # https://docs.python.org/2.5/whatsnew/pep-343.html#context-managers
        return False



#!/usr/bin/env python3

import filecmp
import os
import re
import shutil
import subprocess
import tempfile

from setuptools import setup

VERSIONFILE = "px/version.py"


def write_version(version: str, replace: bool = False) -> None:
    """
    Write the given version to px/version.py.

    If px/version.py is missing, the version will always be written.

    If px/version.py is present, the version will only be written if replace is True.
    """
    with tempfile.NamedTemporaryFile(suffix=".py", delete=False) as tmp:
        tmp.write(
            b"# NOTE: Auto generated by update_version_py() in setup.py, no touchie!\n"
        )
        tmp.write(b'VERSION = "%s"\n' % bytearray(version, "utf_8"))

        # Flushing is required for filecmp.cmp() to work (below)
        tmp.flush()

        if not os.path.isfile(VERSIONFILE):
            # No version file found
            print(f"Creating {VERSIONFILE} with version {version}")
            shutil.move(tmp.name, VERSIONFILE)
        elif replace and not filecmp.cmp(tmp.name, VERSIONFILE):
            print(f"Updating {VERSIONFILE} to new version {version}")
            shutil.move(tmp.name, VERSIONFILE)
        else:
            # VERSIONFILE was already up to date. If we touch it in this case,
            # it will have its file timestamp updated, which will force the slow
            # px_integration_test.py tests to get rerun.
            #
            # Just clean up our tempfile and be merry.
            print(f"Not touching existing {VERSIONFILE}")
            os.remove(tmp.name)


def update_version_py() -> str:
    """
    Update px/version.py with the current git version.

    Returns the version number.
    """
    git_result = subprocess.run(["git", "describe", "--dirty"], capture_output=True)
    if git_result.returncode == 0:
        write_version(git_result.stdout.decode("utf-8").strip(), True)
    else:
        # Don't overwrite any existing version since we just made this one up
        write_version("0.0.0", False)

    from px import version

    return version.VERSION


if __name__ == "__main__":
    version_for_setuptools = update_version_py()

    if not re.match(r"^[0-9]+\.[0-9]+\.[0-9]+$", version_for_setuptools):
        # Setuptools requires a nice version number
        version_for_setuptools = "0.0.0"

    with open(
        os.path.join(os.path.dirname(__file__), "README.rst"), encoding="utf-8"
    ) as fp:
        LONG_DESCRIPTION = fp.read()

    setup(
        name="pxpx",
        version=version_for_setuptools,
        description="ps and top for Human Beings",
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/x-rst",
        author="Johan Walles",
        author_email="johan.walles@gmail.com",
        url="https://github.com/walles/px",
        license="MIT",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "Intended Audience :: System Administrators",
            "License :: OSI Approved :: MIT License",
            "Operating System :: MacOS",
            "Operating System :: POSIX :: Linux",
            "Programming Language :: Python :: 3",
            "Topic :: System :: Monitoring",
            "Topic :: System :: Systems Administration",
            "Topic :: Utilities",
        ],
        packages=["px"],
        # See: http://setuptools.readthedocs.io/en/latest/setuptools.html#setting-the-zip-safe-flag
        zip_safe=True,
        setup_requires=[
            "pytest-runner",
        ],
        entry_points={
            "console_scripts": [
                "px = px.px:main",
                "ptop = px.px:main",
                "pxtree = px.px:main",
            ],
        },
        # Note that we're by design *not* installing man pages here.
        # Using "data_files=" only puts the man pages in the egg file,
        # and installing that egg doesn't put them on the destination
        # system.
        #
        # After trying to figure this out for a bit, my conclusion is
        # that "pip install" simply isn't meant for installing any man
        # pages.
        #
        #   /johan.walles@gmail.com 2018aug27
    )



from . import px_terminal

from . import px_process
from typing import List
from typing import Tuple
from typing import Dict
from typing import Optional


def render_launch_tuple(launch_tuple: Tuple[str, int]) -> str:
    binary = launch_tuple[0]
    count = launch_tuple[1]
    if count == 0:
        return binary

    return px_terminal.bold(binary) + "(" + str(count) + ")"


def _get_minus_max_score(tuples_list: List[Tuple[str, int]]) -> int:
    max_score = 0
    for current_tuple in tuples_list:
        max_score = max(max_score, current_tuple[1])
    return -max_score


def sort_launchers_list(
    launchers_list: List[List[Tuple[str, int]]],
) -> List[List[Tuple[str, int]]]:
    return sorted(launchers_list, key=_get_minus_max_score)


def _strip_parentheses(s: str) -> str:
    if not s:
        return s
    if s[0] != "(":
        return s
    if s[-1] != ")":
        return s
    return s[1:-1]


def _list_new_launches(
    before: List[px_process.PxProcess],
    after: List[px_process.PxProcess],
) -> List[px_process.PxProcess]:
    pid2oldProc: Dict[int, px_process.PxProcess] = {}
    for old_proc in before:
        pid2oldProc[old_proc.pid] = old_proc

    new_procs = []  # List[px_process.PxProcess]
    for new_proc in after:
        if new_proc.pid not in pid2oldProc:
            # This is a new process
            new_procs.append(new_proc)
            continue

        old_proc = pid2oldProc[new_proc.pid]
        if old_proc.start_time != new_proc.start_time:
            # This is a new process, PID has been reused
            new_procs.append(new_proc)
            continue

    return new_procs


def _to_tuple_list(launcher_list: Tuple[str, ...], count: int) -> List[Tuple[str, int]]:
    """
    Converts from: (("a", "b", "c"), 5)
                to  : [("a", 0), ("b", 0), ("c", 5)]
    """
    tuple_list: List[Tuple[str, int]] = []
    for launcher in launcher_list:
        tuple_list.append((launcher, 0))
    tuple_list[-1] = (launcher_list[-1], count)
    return tuple_list


def _merge_tuple_lists(
    tl1: List[Tuple[str, int]],
    tl2: List[Tuple[str, int]],
) -> Optional[List[Tuple[str, int]]]:
    if len(tl1) > len(tl2):
        longer = tl1
        shorter = tl2
    else:
        longer = tl2
        shorter = tl1

    merged = longer[:]
    for i, t1 in enumerate(shorter):
        t2 = longer[i]
        if t1[0] != t2[0]:
            # Mismatch, we can't merge these
            return None
        merged[i] = (t1[0], t1[1] + t2[1])

    return merged


def _callchain(process: px_process.PxProcess) -> Tuple[str, ...]:
    reverse_callchain: List[str] = []

    current: Optional[px_process.PxProcess] = process
    while current is not None:
        reverse_callchain.append(_strip_parentheses(current.command))
        current = current.parent

    return tuple(reversed(reverse_callchain))


class Launchcounter:
    def __init__(self):
        self._hierarchies: Dict[Tuple[str, ...], int] = {}

        # Most recent process snapshot
        self._last_processlist: Optional[List[px_process.PxProcess]] = None

    def _register_launches(self, new_processes: List[px_process.PxProcess]) -> None:
        for new_process in new_processes:
            callchain = _callchain(new_process)
            if callchain in self._hierarchies:
                self._hierarchies[callchain] += 1
            else:
                self._hierarchies[callchain] = 1

    def update(self, procs_snapshot: List[px_process.PxProcess]) -> None:
        if self._last_processlist is None:
            self._last_processlist = procs_snapshot
            return

        new_processes = _list_new_launches(self._last_processlist, procs_snapshot)
        self._register_launches(new_processes)

        self._last_processlist = procs_snapshot

    def _coalesce_launchers(self) -> List[List[Tuple[str, int]]]:
        coalesced: List[List[Tuple[str, int]]] = []

        for launcher_list in sorted(self._hierarchies.keys()):
            count = self._hierarchies[launcher_list]

            new_tuple_list = _to_tuple_list(launcher_list, count)
            if len(coalesced) == 0:
                coalesced.append(new_tuple_list)
                continue

            merged = _merge_tuple_lists(coalesced[-1], new_tuple_list)
            if merged:
                coalesced[-1] = merged
            else:
                coalesced.append(new_tuple_list)

        return coalesced

    def get_screen_lines(self) -> List[str]:
        launchers_list = self._coalesce_launchers()
        launchers_list = sort_launchers_list(launchers_list)

        lines: List[str] = []
        for row in launchers_list:
            line = " -> ".join(map(render_launch_tuple, row))
            lines.append(line)

        return lines



from . import px_process
from . import px_terminal

import sys

from typing import Set, List, Optional, Iterable


def tree(search: str) -> None:
    """Print a process tree"""
    print_me = _generate_tree(px_process.get_all(), search)
    if not print_me:
        print(f"No processes found matching <{search}>", file=sys.stderr)
        return

    for line in print_me:
        print(line)


def _generate_tree(processes: List[px_process.PxProcess], search: str) -> List[str]:
    if not processes:
        return []

    # Only print subtrees needed for showing all search hits and their children.
    #
    # We do that by starting at the search hits and walking up the tree from all
    # of them, collecting each PID we see along the way. Then when we render the
    # tree, we only render those PIDs.
    show_pids: Set[int] = set()
    if search:
        for process in processes:
            if not process.match(search):
                continue

            _mark_children(process, show_pids)

            # Walk up the tree from this process, marking each process for display
            if process and process.parent:
                process = process.parent
            while process:
                if process.pid in show_pids:
                    break
                show_pids.add(process.pid)
                if not process.parent:
                    break
                process = process.parent

        # Search found nothing
        if not show_pids:
            return []

    lines = []
    coalescer = Coalescer(0)
    lines += coalescer.submit(processes[0], search)
    lines += coalescer.flush()
    return lines + _generate_child_tree(processes[0].children, 1, search, show_pids)


def _mark_children(process: px_process.PxProcess, show_pids: Set[int]) -> None:
    """Recursively mark all children of a search hit as needing to be shown"""
    show_pids.add(process.pid)
    for child in process.children:
        _mark_children(child, show_pids)


class Coalescer:
    def __init__(self, indent: int) -> None:
        self._base: Optional[px_process.PxProcess] = None
        self._count = 0
        self._indent = indent

    def submit(self, process: px_process.PxProcess, search: str) -> List[str]:
        """Returns an array of zero or more lines to be printed"""
        is_search_hit = search and process.match(search)
        has_children = bool(process.children)
        is_candidate = not is_search_hit and not has_children

        # If we can coalesce this, do it!
        if self._base and is_candidate and self._base.command == process.command:
            self._count += 1
            return []
        if not self._base and is_candidate:
            self._base = process
            self._count = 1
            return []

        return_me = []

        # Otherwise, print the coalesced line if we have one
        if self._base:
            return_me += self.flush()

        # Can we coalesce this line?
        if is_candidate:
            self._base = process
            self._count = 1
            return return_me

        # And print the current line
        if is_search_hit:
            return_me.append(
                f"{'  ' * self._indent}{px_terminal.bold(process.command)}({process.pid})"
            )
        else:
            return_me.append(f"{'  ' * self._indent}{process.command}({process.pid})")

        return return_me

    def flush(self) -> List[str]:
        if not self._base:
            return []

        assert self._count > 0

        return_me: str
        if self._count == 1:
            return_me = f"{'  ' * self._indent}{self._base.command}({self._base.pid})"
        else:
            return_me = f"{'  ' * self._indent}{self._base.command}... ({px_terminal.bold(f'{self._count}×')})"

        self._base = None
        self._count = 0

        return [return_me]


def _generate_child_tree(
    children: Iterable[px_process.PxProcess],
    indent: int,
    search: str,
    show_pids: Set[int],
) -> List[str]:
    lines = []

    coalescer = Coalescer(indent)
    for child in sorted(
        children, key=lambda p: (p.command.lower(), bool(p.children), p.pid)
    ):
        if show_pids and child.pid not in show_pids:
            continue

        lines += coalescer.submit(child, search)
        lines += _generate_child_tree(child.children, indent + 1, search, show_pids)

    lines += coalescer.flush()
    return lines
